#
# action.tcl - PGBrowse Action procedures
#
# Actions are user-defined Tcl scripts.  See $PGBROWSE/actions/README
# for instructions on adding your own action.
#
# $Header: /nfs/home/src/base/local/tk/pgb3.0/lib/RCS/action.tcl,v 1.2 1993/12/28 21:17:13 jimbo Exp jimbo $
#


#
# ActionInit - setup Actions by scanning site directories
#
proc ActionInit {} {
    global PGBAc PGB env

    # look for user specified directories
    if [info exists env(PGBACTIONS)] {
        lappend sites [split $env(PGBACTIONS) :]
    }

    # look for the standard actions in PGB(Home)
    foreach dir [glob -nocomplain $PGB(Home)/actions/*/] {
        if [info exists sites] {if {[lsearch $sites $dir] != -1} continue}
	lappend sites $dir
    }

    if ![info exists sites] return;    # no site directories
    foreach site $sites {
        if {![file isdirectory $site] || ![file readable $site]} continue
        foreach actiondir [glob -nocomplain $site/*/] {
	    regsub {/$} $actiondir "" actiondir; # remove trailing /
    	    set init $actiondir/$PGB(actionInit)
            if ![file readable $init] continue;  # no init
	    # the init file must set the Name and optionally
	    # the Type and Help variables and must have a procedure
	    # called Action which accepts a single argument, the
	    # action directory.
            set Name ""
            set Help ""
            set Type ""
            source $init
            set PGBAc($actiondir) [list $Name $Help $Type]
        }
    }
}


#
# ActionButton - adds an action button to a Tku_MenuBar menu
#
# input:
#    w    Tku_MenuBar frame
#    type action type (default: all actions)
#
proc ActionButton {menubar {type ""}} {
    global PGB PGBAc

    if ![info exists PGBAc] ActionInit; # scan for all actions


    set lastbutton ""
    set c -1
    foreach actiondir [lsort [array names PGBAc]] {

	# make sure this action is of the correct type
	set Type [lindex $PGBAc($actiondir) 2]
	if {[lsearch $Type $type] == -1 && $type != ""} continue

	# add the menu button now that we have at least one valid action
	if {$c == -1} {
    	    set menu "[Tku_MenuButton $menubar {Actions -under 0}].m"
	}

	# if this is a new site, create a new cascade entry
	set button [file tail [file dirname $actiondir]]
	if {$button != $lastbutton} {
	    incr c
	    set i 0
	    $menu add cascade -label $button -menu $menu.$c
	    menu $menu.$c
	    set lastbutton $button
	}

	# add the entry and it's popup help
	$menu.$c add command -label [lindex $PGBAc($actiondir) 0] \
		    -command "ActionExec $actiondir"
	if {[lindex $PGBAc($actiondir) 1] != ""} {
	    Tku_Popup [lindex $PGBAc($actiondir) 1] $menu.$c $i
	}
	incr i
    }
}

#
# ActionExec -    execute an action
#
# Input:
#	actiondir	the action directory, passed to the Action
#
proc ActionExec {actiondir} {
    global PGB

    set init $actiondir/$PGB(actionInit)
    if {![file readable $init]} {
        Tku_Error "Can't run action in directory:  $actiondir"
        return 0
    }
    # remove current Action
    proc Action {} {}
    # source new Action
    source $init
    if {[info body Action] == {}} {
        Tku_Error "No Action procedure defined in $init, can't run action."
        return 0
    }
    # run Action
    Action $actiondir
}

