#
# browse.tcl - procedures for the top level window of PGBrowse
#
# $Header: /nfs/home/src/base/local/tk/pgb3.0/lib/RCS/browse.tcl,v 1.3 1993/12/30 01:10:35 jimbo Exp jimbo $
#

# ---------------------------------------------------------------------------
# NewPGBrowser - Open new browser
# ---------------------------------------------------------------------------
proc File:NewBrowser {} {
    global PGB

    set cmd "$PGB(Home)/bin/pgbrowse -host [PQ host] -port [PQ port]"

    # add name of database and open classes if any
    if {[PQ db] != ""} {append cmd " -db [PQ db]"}
    foreach class [Class opened] {lappend classes $class}
    if [info exists classes] {append cmd " -class [join $classes ,]"}

    # add top level options
    foreach opt {debug system verify} {
	if $PGB(-$opt) {append cmd " -$opt"}
    }

    # exec the new browser
    Tku_Working "Starting new browser..."
    eval exec $cmd &
    # wait a bit for the browsers to start...
    after 2000
    Tku_Waiting
}

# ---------------------------------------------------------------------------
# CloseBrowser - close the browser
# ---------------------------------------------------------------------------
proc CloseBrowser {} {exit}

# ---------------------------------------------------------------------------
# ClearClass - deletes all instances in the current database
# ---------------------------------------------------------------------------
proc ClearClass {} {

    if {[set class [SelectClass "DELETE all instances from class:"]] == ""} return
    set n [Count $class]
    case $n in {
	0 {Tku_Notice "There are no instances in $class."; return}
	-1 {Tku_Error "Could not get count of instances in $class.";return}
    }
    if ![Tku_Confirm "Delete all $n instances from class $class?" "Yes, delete all" "No, keep"] return
    if ![PQ exec "delete $class" "Deleting..."] {
    	PQ errbox "DELETE all failed"
    	return 0
    }
    return 1
}

# ---------------------------------------------------------------------------
# DestroyClass - destroy the currently selected class
# ---------------------------------------------------------------------------
proc DestroyClass {} {

    if {[set class [SelectClass "DESTORY class:"]] == ""} return
    if ![Tku_Confirm "Destroy class $class in database [PQ db]?" "Yes, destroy" "No, keep"] return

    PGBSupportDelete $class

    if ![PQ exec "destroy $class" "Destroying..."] {
         PQ errbox "DESTROY class failed"
         return 0
    }

    GetClasses [PQ db]
    return 1
}

# ---------------------------------------------------------------------------
# CreateDB -    create a new database on the current host
# ---------------------------------------------------------------------------
proc CreateDB {} {
    if {[set db [Tku_Entry {New Database} {} {Create New Database}]] == {}} return

    # make sure the db is a valid char16
    if {[string length $db] > 16} {
        Tku_Error "Invalid Name: \"$db\".  Database names must be <= 16 characters in length."
	return
    }
    if [regexp {^[^a-zA-Z]|\ } $db]  {
        Tku_Error "Invalid Name: \"$db\".  Database names must begin with an alphabetic character and contain no white space."
        return
    }

    if {[PQ db] == ""} {PQ setdb template1}
    
    # attempt to add the database and update the database list
    if ![PQ exec "createdb $db" "Creating $db..."] {
        PQ errbox "CREATEDB $db failed"
        return 0
    } else {
    	GetDbs [PQ host] [PQ port]
    	GetClasses $db
    	return 1
    }
}

# ---------------------------------------------------------------------------
# SetPGHost -	prompt for a new pghost
# ---------------------------------------------------------------------------
proc SetPGHost {} {
    if {[set host [Tku_Entry "Postgres Server" [PQ host] "Set PGHOST"]] == {}} return
    GetDbs $host [PQ port]
}

# ---------------------------------------------------------------------------
# File:SetPGPort -	prompt for a new pgport
# ---------------------------------------------------------------------------
proc SetPGPort {} {
    if {[set port [Tku_Entry "Postgres Port" [PQ port] "Set PGPORT"]] == {}} return
    GetDbs [PQ host] $port
}

# ---------------------------------------------------------------------------
# GetDbs - get a list of databases on the host, using the template1
#	database as a start, and update the .dbs.listbox
# ---------------------------------------------------------------------------
proc GetDbs {host port} {
    global PGB

    # connect to the database and clear out the dbs and classes boxes
    Tku_Working "Connecting to $host..." 750
    PQ sethost $host
    PQ setport $port

    if [winfo exists .dbs] {
	Tku_ListboxClear .dbs.listbox
	.dbs.title conf -text $host:$port
    }

    Class reset ""

    # retrieve and save the list of databases using template1
    PQ setdb template1
    if ![PQ exec "retrieve (pg_database.datname)" "Getting databases..."] {
    	PQ errbox "Could not get database list"
    	return 0
    }
    set PGB(dbs) [lsort [PQ getcol 0]]
    PQ setdb ""

    if [winfo exists .dbs] {Tku_ListboxFill .dbs.listbox $PGB(dbs)}
    Tku_DefFocus .dbs.listbox
}


# ---------------------------------------------------------------------------
# ShowCount -	Popup number of instances in a class
# ---------------------------------------------------------------------------
proc ShowCount {} {
    if {[set class [SelectClass "Count instances in class:"]] == ""} return
    if {[set n [Count $class]] == -1} {
    	PQ errbox "Could not get count of instances"
    } else {
	if {$n == 1} {Tku_Notice "$n instance."} \
	else {Tku_Notice "$n instances."}
    }
}

# ---------------------------------------------------------------------------
# Count - count the instances in a class
# ---------------------------------------------------------------------------
proc Count {class} {
    if ![PQ exec "retrieve (instances = count{$class.oid})"] {
    	return -1
    } else {
    	return [PQ get 0 0]
    }
}

# ---------------------------------------------------------------------------
# SelectClass - have the user select a class 
# ---------------------------------------------------------------------------
proc SelectClass {msg} {
    if ![CheckDB] {return ""}

    set classes [Class opened]
    if {[set cur [.classes.listbox cur]] != ""} {
	foreach i $cur {
	    set s [.classes.listbox get $i]
	    if {[lsearch $classes $s] == -1} {lappend classes $s}
	}
    }

    case [llength $classes] in {
	0 {Tku_Notice "Please open or select a class."; return ""}
	1 {return [lindex $classes 0]}
	default {return [Tku_Select $msg $classes]}
    }
}

# ---------------------------------------------------------------------------
# CheckDB - check that a database has been selected
# ---------------------------------------------------------------------------
proc CheckDB {} {
    if {[PQ db] != ""} {
	return 1
    } else {
        Tku_Notice "Please open a database."
    	return 0
    }
}

# ---------------------------------------------------------------------------
# DebugReSource - re-source a file in the $PGB(Home)/lib directory
# ---------------------------------------------------------------------------
proc DebugReSource { file } { 
    global PGB
    set f $PGB(Home)/lib/$file
    if [file exists $f] {
    	puts stderr "re-reading $f..."
    	uplevel #0 "source $f"
    }
}

# ---------------------------------------------------------------------------
# PQ - libpq interface procedure
# ---------------------------------------------------------------------------
proc PQ {cmd {a1 ""} {a2 ""}} {
    global PQ PGB

    # This procedure calls various libpq commands which are compiled
    # into the wish interpreter and maintains a few global variables
    # to keep track of the host/port/db/class.  PQ always assumes
    # just a single group is returned in the portal


    case $cmd in {
	host {PQhost}
	port {PQport}
	db {PQdb}
	dbs {return [Tku_ListboxDump .dbs.listbox]}
        sethost {PQfinish; PQsethost $a1}
        setport {PQfinish; PQsetport $a1}
        setdb {PQsetdb $a1}
	export {	# export the host/port/db to the environment
	    global env
	    set env(PGHOST) [PQ host]
	    set env(PGPORT) [PQ port]
	    if {[PQ db] != ""} {
		set env(DATABASE) [PQ db]
		set env(PGDATABASE) [PQ db]
	    } else {
	    	catch {unset env(DATABASE)}
	    	catch {unset env(PGDATABASE)}
	    }
	}
        ncols {return $PQ(ncols)}
        nlines {return $PQ(nlines)}
	getcol {	# get a whole column as a list
	    if !$PQ(nlines) {return ""}
	    for {set n 0} {$n < $PQ(nlines)} {incr n} {
		lappend col [PQgetvalue $PQ(pbuff) $n $a1]
	    }
	    return $col
	}
        getline {	# get a whole line as a list
	    if !$PQ(nlines) {return ""}
	    for {set n 0} {$n < $PQ(ncols)} {incr n} {
		lappend line [PQgetvalue $PQ(pbuff) $a1 $n]
	    }
	    return $line
	}
        get {		# get a single element
	    if !$PQ(nlines) {return ""} \
	    else {return [PQgetvalue $PQ(pbuff) $a2 $a1]}
	}
        getname {	# get a column name
	    return [PQfnameGroup $PQ(pbuff) 0 $a1]
	}
        getnames {	# return a list of all column names
            for {set n 0} {$n < [PQ ncols]} {incr n} {
                lappend r [PQ getname $n]
            }
            return $r
        }

        status {return $PQ(status)}
        errmsg {return $PQ(errmsg)}
        errbox {	# popup a return code of the last query
	    case $PQ(status) in {
	    0 	{Tku_Error "$a1\n\nPostgres error:\n\n$PQ(errmsg)"}
	    1	{Tku_Notice "$a1\n\nPostquel successful, $PQ(nlines) lines returned."}
	    2	{Tku_Notice "$a1\n\nPostquel successful, nothing returned."}
	    3	{Tku_Error "$a1\n\nPostquel cancelled."}
	    4	{Tku_Error "$a1\n\nNothing to send to the backend."}
	    default {Tku_Error "$a1\n\nUnknown status: $PQ(status)"}
	    }
	}
        exec {		# PQExec a query
	    PQDoExec $a1 $a2
	}
        reset {catch {unset PQ}}
        default {error "PQ: unknown command:  $cmd"}
    }
}

# ---------------------------------------------------------------------------
# PQDoExec - execute a Postquel query
# ---------------------------------------------------------------------------
proc PQDoExec {cmd msg} {
    global PGB PQ

    PQ reset
    if {$msg == ""} {set msg "Querying..."}

    # trim and/or allow the user to edit the query
    set cmd [string trim $cmd]
    if $PGB(-verify) {
	if ![Tku_Edit "Modify and/or Confirm Postquel" $cmd cmd] {
	    set PQ(status) 3
	    return 0
	}
    	set cmd [string trim $cmd]
	if {$cmd == ""} {
	    set PQ(status) 4
	    return 0
	}
    }

    # log and send the query
    Log $cmd "Postquel"
    Tku_Working $msg
    set res [PQexec $cmd]
    Tku_Waiting

    # log the query return code
    case [string range $res 0 0] in {
	{R E} {
	    Log "Postquel error:  [set PQ(errmsg) $res]" Error
	    return [set PQ(status) 0]
	}
	C {
	    Log $res Success
	    return [set PQ(status) 1]
	}
	P {
    	    set PQ(pbuff) [PQparray [string range $res 1 end]]
    	    set PQ(nlines) [PQntuplesGroup $PQ(pbuff) 0]
    	    set PQ(ncols) [PQnfieldsGroup $PQ(pbuff) 0]
	    Log "$res - $PQ(ncols) columns, $PQ(nlines) lines" Success
            	    return [set PQ(status) 1]
	}
	default {
	    set PQ(errmsg) "Odd return code: $res"
	    return [set PQ(status) 0]
	}
    }
}

# ---------------------------------------------------------------------------
# Class - open, close, and maintain information about classes in the db
# ---------------------------------------------------------------------------
proc Class {cmd {arg ""}} {
    global PGB PGBC

    if ![info exists PGBC] {set PGBC(num) 0}

    case $cmd in {
	num {return $PGBC(num)}
	classes {
	    if $PGBC(num) {return $PGBC(classes)} \
	    else {return ""}
	}
	numopened {return [llength [Class opened]]}
	opened {
	    if ![info exists PGBC(opened)] {return ""} \
	    else {return $PGBC(opened)}
	}
	open {
	    if {$arg == ""} {error "Class:  must provide a class to open."}
	    if [info exists PGBC(n,$arg)] {
		wm iconify .class$PGBC(n,$arg)
		wm deiconify .class$PGBC(n,$arg)
		update
		focus .class$PGBC(n,$arg)
		return
	    }
	    if {[lsearch [Class classes] $arg] == -1} {
		Tku_Error "No such class:  $arg"
		return 0
	    }
	    lappend PGBC(opened) $arg
	    # get the next available window name
	    set n 0
	    while {[winfo exists .class$n]} {incr n}
	    set PGBC(n,$arg) $n
	    set f .class$n
	    toplevel $f
	    wm title $f "Class: $arg"
	    wm iconname $f $arg
	    wm group . $f

    	    Tku_Listbox $f.box Query -geom 30x10
	    $f.box.title conf -text $arg
    	    pack $f.box -expand 1 -fill both
	    Tku_Button $f.query Query "Query $arg" \
		"Open the query tool on $arg" 1
	    Tku_Button $f.close Close "destroy .class$n" ""
	    bind $f <Destroy> "Class close $arg"
	    pack $f.query $f.close -pady 2m -padx 5m -side left
	    foreach name [Attribute names $arg] {
		lappend defs [Attribute def $arg $name]
	    }
	    Tku_ListboxFill $f.box.listbox $defs
	    bind $f.box.listbox <Control-w> "$f.close.b invoke"
	    bind $f.box.listbox <Control-q> "$f.query.b invoke"
	    Tku_SetTabbing $f.box.listbox $f.query.b $f.close.b
	    update
	    focus $f.box.listbox
	    return
	}
	close {
	     set PGBC(opened) [ldelete $PGBC(opened) $arg]
	     unset PGBC(n,$arg)
	}
    	reset {
	    if [info exists PGBC] {
		# close any query tool
		if [winfo exists .q] QueryClose
		# delete all possible open class windows
		set n 0
	    	foreach c [Class classes] {
		    if [winfo exists .class$n] {destroy .class$n}
		    incr n
	    	}
		# clear the class and attribute information
		unset PGBC
		if [winfo exists .classes.listbox] {
		    Tku_ListboxClear .classes.listbox
		    .classes.title conf -text $arg
		}
		Attribute reset
		PQ setdb $arg
	    }
	}
    }
}
	
# ---------------------------------------------------------------------------
# GetClasses - query the db for a list of classes
# ---------------------------------------------------------------------------
proc GetClasses {{db ""}} {
    global PGB PGBC

    if {$db == ""} {if {[set db [PQ db]] == ""} return}

    # check for a ModClass in progress
    if [winfo exists .class] {
	Tku_Notice "A class is being modified in the [PQ db] database."
	return 0
    }

    # confirm the db exists
    if {[lsearch $PGB(dbs) $db] == -1} {
	Tku_Error "No such database:  $db"
	return 0
    }
    
    Tku_Working "Opening $db..." 750

    # clear out the class arrays and update the .classes listbox
    Class reset $db

    # check for the pgb_support class and create it if it doesn't exist
    if ![VerifyPGBSupport] {
	Tku_Error "The pgb_support class does not exist or could not be created - please check to log."
	Tku_Waiting
	return 0
    }

    # send the Postquel 
    set cmd {
retrieve (
    pg_class.relname
)
    where pg_class.relkind = "r"	/* skip indexes and other stuff */
    and	  pg_class.relname !~ "^Xinv"	/* skip large objects */}

    if ![PQ exec $cmd "Getting classes..."] {
    	PQ errbox "Could not get classes list"
	Tku_Waiting
    	return 0
    }

    set PGBC(num) [PQ nlines]
    set PGBC(classes) [lsort [PQ getcol 0]]
    FillClasses
}

proc FillClasses {} {
    global PGB PGBC

    # update the .classes listbox
    if ![winfo exists .classes] {return}
    Tku_ListboxClear .classes.listbox
    if ![info exists PGBC] {return}
    foreach c $PGBC(classes) {
	if [regexp {^pg_} $c] {
		lappend sys $c
	} else {
		lappend classes $c
	}
    }
    if $PGB(-system) {lvarcat classes $sys}
    Tku_ListboxFill .classes.listbox $classes
    Tku_DefFocus .classes.listbox
    return 1
}

# ---------------------------------------------------------------------------
# Attribute - information about the attributes of classes in the db
# ---------------------------------------------------------------------------
proc Attribute {cmd {class ""} {arg ""} {arg2 ""}} {
    global PGBA PGB

    # get the attributes of a class if we haven't already done so
    if {$class != ""} {
	if ![info exists PGBA($class,num)] {
	    if ![AttributeQuery $class] return 
	}
    }

    case $cmd in {
	reset	{catch {unset PGBA}}
	resetmeta {AttributeResetMeta $class $arg $arg2}
	num	{return $PGBA($class,num)}
	name	{return $PGBA($class,name,$arg)}
	type	{return $PGBA($class,type,$arg)}
	meta	{return $PGBA($class,meta,$arg)}
	dim	{return $PGBA($class,dim,$arg)}
	com	{return $PGBA($class,comm,$arg)}
	length {return $PGBA($class,length,$arg)}
	retrieve {return $PGBA($class,retrieve,$arg)}
	choice	{return $PGBA($class,choices,$arg)}
	maxline {return $PGBA($class,maxline)}
        fulltype {return $PGBA($class,ftype,$arg)}
        def {return $PGBA($class,def,$arg)}
        names {return $PGBA($class,names)}
        default {
            error "Attribute error: no such command $cmd"
        }
    }
}

# ---------------------------------------------------------------------------
# AttributeQuery - get the attributes of a class
# ---------------------------------------------------------------------------
proc AttributeQuery {class} {
    global PGB PGBA

    set PGBA($class,maxline) 0

    set cmd "
retrieve (
    pg_attribute.attname, pg_attribute.attnum,
    pg_type.typname, pg_attribute.attnelems,
    metadata=pgb_get_meta(\"$class\"::char16,pg_attribute.attname)
)
    where pg_attribute.attrelid = pg_class.oid
    and   pg_attribute.atttypid = pg_type.oid
    and   pg_attribute.attnum   > 0
    and   pg_class.relname      = \"$class\"
"
    set msg "Opening $class..."
    if ![PQ exec $cmd $msg] {
	PQ errbox "Could not get attributes for class:  $class"
    	return 0
    }

    Tku_Working $msg

    lappend PGBA(classes) $class
    set PGBA($class,num) [PQ nlines]

    for {set n 0} {$n <= $PGBA($class,num)} {incr n} {
	if {$n == $PGBA($class,num)} {
	    lappend PGBA($class,names) [set name oid]
	    set num $PGBA($class,num)
	    set type oid
	    set dim 0
	    set meta "Object ID"
	} else {
	    lappend PGBA($class,names) [set name [PQ get 0 $n]]
 	    set num [expr [PQ get 1 $n] - 1]
	    set type [PQ get 2 $n]
            regsub {^_} $type {} type
 	    set dim [PQ get 3 $n]
	    if {$dim == 1} {set dim 0}
	    set meta [PQ get 4 $n]
	}

       	set PGBA($class,name,$num) $name
       	set PGBA($class,type,$name) $type
       	set PGBA($class,dim,$name) $dim
       	set PGBA($class,meta,$name) $meta
	case $dim in {
	    -1 {set PGBA($class,ftype,$name) "$type\[\]"}
	    {1 0} {set PGBA($class,ftype,$name) $type}
	    default {
		set PGBA($class,ftype,$name) "$type\[$dim\]"
	    }
	}

	AttributeResetMeta $class $name $meta
    }
    incr PGBA($class,num)

    Tku_Waiting

    return 1
}


proc AttributeResetMeta {class name meta} {
	global PGBA

	set comm ""
	set choices ""
	set retrieve 1
	set length 20
	if {$meta != "(null)"} {
	    set meta [split $meta :]
	    set comm [lindex $meta 0]

	    foreach c [lrange $meta 1 end] {
		case $c in {
		    length=* {set length [lindex [split $c =] 1]}
		    retrieve=* {set retrieve [lindex [split $c =] 1]}
		    default {lappend choices $c}
		}
	    }
	    if ![info exists choices] {
                if {$type == "bool"} {
                    set choices [list TRUE FALSE]
	        } else {
	            set choices ""
	        }
	    }
	}
      	set PGBA($class,comm,$name) $comm
       	set PGBA($class,choices,$name) $choices
       	set PGBA($class,retrieve,$name) $retrieve
       	set PGBA($class,length,$name) $length
        if {$comm == ""} {
	    set PGBA($class,def,$name) "$name ($PGBA($class,type,$name))"
	} else {
	    set PGBA($class,def,$name) "$name ($PGBA($class,type,$name): $comm)"
	}
    	set l [string length $PGBA($class,def,$name)]
    	if {$l > $PGBA($class,maxline)} {set PGBA($class,maxline) $l}
}

proc File:Open {} {
    global PGB

    if {[.dbs.listbox cur] != ""} {
	GetClasses [selection get]
    } elseif {[.classes.listbox cur] != ""} {
	puts stdout [selection get]
	foreach class [selection get] {Class open $class}
    }
}

proc Debug:SendTK {} {
    if ![Tku_Edit "Tk To Execute" "" cmd] return
    send [winfo name .] $cmd
}
