
# query.tcl - procedures to support the query tool and query results window
#
# $Header: /nfs/home/src/base/local/tk/pgb3.0/lib/RCS/query.tcl,v 1.4 1993/12/30 01:10:44 jimbo Exp jimbo $
#

# ---------------------------------------------------------------------------
# Query - Simple query system
# ---------------------------------------------------------------------------
proc Query {{classes ""}} {
    global PGB PGBOpers

    if ![CheckDB] return

    if {$classes == ""} {
	if {[set classes [Class opened]] == ""} {
	    if {[set classes [Tku_ListboxGet .classes.listbox]] == ""} {
	    	if {[set classes [SelectClass "Query class:"]] == ""} return
	    }
	}
    }

    set PGB(queryClasses) $classes

    if {[llength $PGB(queryClasses)] == 1} {
	set PGB(queryUpdatable) 1
    } else {
	set PGB(queryUpdatable) 0
    }

    if [winfo exists .q] {destroy .q}
    toplevel .q
#    Tku_HelpWindow .q
    wm title .q Query
    wm protocol .q WM_DELETE_WINDOW QueryClose
    wm geom .q +150+150
    wm minsize .q 50 50
    wm withdraw .q

    # pack together menu and info bars
    Tku_MenuBar .q.menu \
	{About {This window allows you to query the class(s) you're
	interested in.  First, set "on" the checkbutton's of the fields
	you're interested in retrieving.  Then, either select "Retrieve
	All" to return all rows or enter qualifications for one or more
	fields and select "Look Up."  If you're querying a single class
	you can insert a new row by filling in the fields and selecting
	"Enter".}}

    Tku_MenuButton .q.menu "File -under 0" \
        "command -label {Show Log} -command LogOpen -accel C-L -under 5" \
        {Opens the log window.} \
        "command -label Close -command QueryClose -accel C-W -under 0" \
        {Closes the Query Dialog and Query Results window.}


    Tku_StatusBar .q.info Query

    pack .q.menu .q.info -fill x

    frame .q.buttons1
    # pack together button1 frame
    if $PGB(queryUpdatable) {
        Tku_Button .q.append Append QueryAppend \
	    "APPENDS a new instance to the database"
        pack .q.append -in .q.buttons1 -side right -padx 2m -fill x
    }
    Tku_Button .q.retrieve "Retrieve" QueryRetrieve \
	"RETRIEVE matching instances" 
    Tku_Button .q.retrieveall "Retrieve ALL" QueryRetrieveAll \
	"RETRIEVE all instances"
    pack .q.retrieve .q.retrieveall \
        -in .q.buttons1 -side right -padx 2m -fill x
    #
    # This stuff is for the Sequoia project - it probably won't work for
    # you unless you have IDL and whatever else it needs...
    #
    set PGB(grQuery) 0
    if $PGB(grQuery) {
    	Tku_Button .q.gr_query "Graphical Query" GraphQuery \
	    "Graphical Query using IDL and a cursor to select area of interest"
        pack .q.gr_query -in .q.buttons1 -side right -padx 2m -fill x
	lappend tablist .q.gr_query.b
    }
    lappend tablist .q.retrieveall.b .q.retrieve.b
    if $PGB(queryUpdatable) {
    	lappend tablist .q.append.b
    }

    # pack together button2 frame
    frame .q.buttons2 
    pack [label .q.atts -text Attributes: -anchor w] \
	-side left -in .q.buttons2 
    Tku_Button .q.attsdef Default QueryAttsDef \
	"Select the default set of attributes for retrieval"
    Tku_Button .q.attsall All QueryAttsAll \
	"Select all attributes for retrieval"
    Tku_Button .q.attsnone None QueryAttsNone \
	"Select no attributes for retrieval"
    Tku_Button .q.clear Clear QueryClear \
	"Clear all qualification entry boxes"
    lappend tablist .q.attsdef.b .q.attsall.b .q.attsnone.b .q.clear.b 
    pack .q.attsdef .q.attsall .q.attsnone .q.clear -side left \
	-in .q.buttons2


    pack .q.buttons2 -anchor w -fill x

    # build and pack each label/entry pair; setup <Return> binding

    frame .q.cframe -bd 0
    scrollbar .q.xs -command {.q.c xview} -orient horiz -relief flat
    scrollbar .q.ys -command {.q.c yview} -relief flat
    canvas .q.c -yscroll {.q.ys set} -bd 0 -xscroll {.q.xs set}

    frame .q.c.f
    .q.c create window 0 0 -window .q.c.f -anchor nw

    # size of the choice/operator buttons or blank space
    set PGB(query,butwid) 12
    set PGB(query,butbd) 2
    set PGB(query,space) 20

    set maxline 0
    foreach class $PGB(queryClasses) {
	foreach name [Attribute names $class] {
      	    if {[set n [Attribute maxline $class]]>$maxline} {set maxline $n}
	}
    }

    Tku_Working "Opening Query Tool..."
    set i 0
    # setup the attributes to retrieve
    QueryAttsDef
    foreach class $PGB(queryClasses) {

      # pack a spacer between classes
      label .q.c.f.l$i -text $class: -anchor w 
      pack .q.c.f.l$i -fill x

      foreach name [Attribute names $class] {
	set PGB(query,class,$i) $class
	set PGB(query,name,$i) $name
        set frame [frame .q.c.f.$i -bd 0]
	set type [Attribute type $class $name]

	# add a checkbutton for retrieveing this attribute
	checkbutton $frame.r -var PGB(ret,$class,$name) -relief flat \
	    -anchor w -text [Attribute def $class $name] -width $maxline


	# add an entry pulldown for this attribute

	Tku_PullDown $frame.entry {} {} {} -width 30
        if {[Attribute choice $class $name] != "" || !$PGB(-nomodify)} {
		QueryOptionsPullDown $frame $class $name
	} else {
		$frame.entry.b conf -relief flat -bitmap "" -state disabled
	}
		

	# add an operator pulldown for this attribute
	if ![info exists PGBOpers($type)] {
	    if ![PQ exec "
retrieve (pg_operator.oprname)
where	pg_operator.oprleft = pg_type.oid
and	pg_type.typname = \"$type\"
" "Getting $type operators..."]	    {
		PQ errbox "Can't get operators for type: $type"
		destroy .q
		return
	    }
	    set PGBOpers($type) [PQ getcol 0]
	}
	Tku_PullDown $frame.opts {} $PGBOpers($type) = -width 3

#	Tku_PullDown $frame.andor {} "where and or" where -width 8
#	set PGB(queryAndOr,$i) and
#	radiobutton $frame.and -text where -var PGB(queryAndOr,$i) -val and
#	radiobutton $frame.or -text or -var PGB(queryAndOr,$i) -val or
#	button $frame.next -text Next... -command "QueryNext $frame"
#	pack $frame.r $frame.andor $frame.opts $frame.entry \
#	    $frame.next -side left 
	pack $frame.r $frame.opts $frame.entry -side left 

        lappend tablist $frame.opts.e $frame.entry.e
        if $i {
	    bind .q.c.f.[expr $i-1].entry.e <Return> \
		"focus $frame.entry.e"
	}

	
	lappend sortby $name

	# set follow-the-mouse focus bindings
	set abg [lindex [$frame.r conf -activebackground] 4]
	set bg [lindex [$frame.entry.e conf -bg] 4]

	bind $frame.r <Any-Enter> [bind Checkbutton <Any-Enter>]
	bind $frame.r <Any-Enter> \
	    "+focus $frame.entry.e; $frame.entry.e conf -bg $abg"

	bind $frame.r <Any-Leave> [bind Checkbutton <Any-Leave>]
	bind $frame.r <Any-Leave> "+$frame.entry.e conf -bg $bg"

	bind $frame.entry.e <Any-Enter> \
	    "+focus $frame.entry.e; $frame.entry.e conf -bg $abg; \
	     $frame.r conf -bg $abg"
	bind $frame.entry.e <Any-Leave> \
	    "$frame.entry.e conf -bg $bg; $frame.r conf -bg $bg"

	pack $frame -fill x

        incr i
      }

    }
    set PGB(queryNum) $i

#    Tku_Working "Opening form..."

    pack .q.xs -in .q.cframe -side bottom -fill x
    pack .q.ys -in .q.cframe -fill y -side right
    pack .q.c -in .q.cframe -fill both -expand 1
    pack .q.buttons1 -fill x -side bottom
    pack .q.cframe -fill both -expand 1

    # set <Return> binding for last label/entry pair
    bind .q.c.f.[expr $i-1].entry.e <Return> "focus .q.retrieve.b"

    # finish and add the sort and options menu buttons
    set PGB(querySortBy) -1
    lrmdups sortby
    foreach name $sortby {
	lappend sorts \
            "radiobutton -value $name -var PGB(querySortBy) -label $name" ""
    }
    lappend sorts \
        "radiobutton -value -1 -var PGB(querySortBy) -label {No Sorting}" \
	"Do not sort the output"
    eval Tku_MenuButton .q.menu "{\"Sort By\" -under 0}" $sorts

    # add Options button
    if ![info exists PGB(query,regexp)] {set PGB(query,regexp) 1}
    if ![info exists PGB(query,unique)] {set PGB(query,unique) 0}
    Tku_MenuButton .q.menu "Options -under 0" \
	"checkbutton -label Unique -var PGB(query,unique) -under 0" \
	"Remove duplicate instances from the output." \
	"radiobutton -label {Exact Matching} -var PGB(query,regexp) -val 0 -under 0 -comm QueryRegexpOpt" \
	"Use exact matches by default for text and character comparisions." \
	"radiobutton -label {Regular Expressions} -var PGB(query,regexp) -val 1 -under 0 -comm QueryRegexpOpt" \
	"Use regular expression (or substrings) by default for text and character comparisions." \
	"separator" "" \
	"checkbutton -label {Verify Postquel} -variable PGB(-verify) -under 0" \
	{Modify and/or confirm Postquel before sending to the backend}


    # set tab binding
    eval Tku_SetTabbing $tablist

    # set the default char/text matching
    QueryRegexpOpt

    if $PGB(-hide) {wm withdraw .}


    QueryResize
    set th [set h [winfo reqheight .q.c.f]]
    set maxh [expr [winfo screenheight .q]*.5]
    if {$th > $maxh} {set th $maxh}
    .q.c conf -width [winfo reqwidth .q.c.f] -height $th
#    wm minsize .q [winfo reqwidth .q.c.f] 200
    wm geom .q +10+10
    wm deiconify .q
    Tku_DefFocus .q.c.f.0.entry.e
    Tku_Waiting

}

proc QueryResize {} {
    update
    set reqh [winfo reqheight .q.c.f]
    .q.c conf -scrollregion "0 0 [winfo reqwidth .q.c.f] $reqh"
    set h [lindex [.q.c conf -height] 4]
    if {$h > $reqh} {.q.c conf -height $reqh}
}

proc QueryGetQual {i} {
    global PGB
    set class $PGB(query,class,$i)
    set name $PGB(query,name,$i)
    set type [Attribute type $class $name]
    set dim [Attribute dim $class $name]
    set oper [.q.c.f.$i.opts.e get]
    set data [.q.c.f.$i.entry.e get]
#    set andor [.q.c.f.$i.andor.e get]
#    if {$andor == "where"} {set andor and}
#    set andor $PGB(queryAndOr,$i)
 set andor and
    if {$data == {}} {return ""}
    set join 0
    foreach otherclass $PGB(queryClasses) {
	if [regexp "^$otherclass." $data] {set join 1}
    }
    if !$join {
    	if {$dim == 0 || $dim == 1} {
	    	set data "\"$data\"::$type"
	} else {
	  	foreach e [split $data ,] {lappend d "\"$e\"::$type"}
	 	set data "{[join $d ,]}"
	}
    }
    return "$andor $class.$name $oper $data"
}

proc QueryNext {frame} {
    global PGB
    set i [winfo name $frame]
    set qual [QueryGetQual $i]
    if {$qual == ""} return 
    $frame.entry.e delete 0 end
    set nframe [winfo parent $frame].n$i
    set n 0
    if ![winfo exists $nframe] {
	frame $nframe -bd 0
	pack $nframe -after $frame -fill x
    } else {
	while {[winfo exists $nframe.$n]} {incr n}
    }
    button $nframe.$n -cursor pirate -text $qual \
	-bd 0 -command "QueryNext:Destroy $nframe $n" -anchor e
    pack $nframe.$n -padx 5m -anchor e
    QueryResize
}

proc QueryNext:Destroy {nframe n} {
    destroy $nframe.$n
    if {[winfo children $nframe] == ""} {destroy $nframe}
    QueryResize
}

proc QS:Delete {} {
    if {[set cur [.q.qs.listbox cur]] == ""} return
    foreach i $cur {.q.qs.listbox delete $i}
}

proc QueryOptionsPullDown {frame class name} {
    global PGB

    set choices [Attribute choice $class $name]
    set type [Attribute type $class $name]

    set pd $frame.entry
    $pd.b.m delete 0 last
    foreach c $choices {
	$pd.b.m add command -label $c -command \
	    "$pd.e delete 0 end; $pd.e insert 0 {$c}"

    }

    if !$PGB(-nomodify) {
	if {[$pd.b.m index last] != "none"} {$pd.b.m add separator}
	$pd.b.m add command -label "Options..." \
	    -command "QueryModOptions $frame $class $name"
    }
}

# ---------------------------------------------------------------------------
# QueryRegexpOpt - set the default search type for char and text types
# ---------------------------------------------------------------------------
proc QueryRegexpOpt {} {
    global PGB

    if $PGB(query,regexp) {
	set right ~
	set wrong =
    } else {
	set right =
	set wrong ~
    }

    set i 0
    foreach class $PGB(queryClasses) {
	foreach name [Attribute names $class] {
	    set type [Attribute type $class $name]
	    if {[string match *text $type] || [string match *char* $type]} {
		if {[.q.c.f.$i.opts.e get] == $wrong} {
		    .q.c.f.$i.opts.e delete 0 end
		    .q.c.f.$i.opts.e insert 0 $right
		}
	    }
	incr i
	}
    }
}
# ---------------------------------------------------------------------------
# QueryRet* - set the attributes to retrieve
# ---------------------------------------------------------------------------
proc QueryAttsDef {} {
    global PGB
    foreach class $PGB(queryClasses) {
	foreach name [Attribute names $class] {
	    set PGB(ret,$class,$name) [Attribute retrieve $class $name]
	}
    }
}
proc QueryAttsNone {} {QueryAttsAll 0}
proc QueryAttsAll {{allnone 1}} {
    global PGB 
    foreach class $PGB(queryClasses) {
    	foreach name [Attribute names $class] \
  	    {set PGB(ret,$class,$name) $allnone}
    }
}

# ---------------------------------------------------------------------------
# QueryModOptions - modify attribute options (length, retrieve, choices)
# ---------------------------------------------------------------------------
proc QueryModOptions {frame class name} {
    global PGB

    catch {destroy .qmod}
    toplevel .qmod
    wm title .qmod "Modify Options for $name"
    wm iconname .qmod "Options: $name"
    frame .qmod.top
    Tku_EntryBox .qmod.com Comment: [Attribute com $class $name] -width 40
    loop n 0 100 10 {lappend ls $n}
    Tku_PullDown .qmod.l "Column Length:" $ls \
	[Attribute length $class $name] -width 4
    set PGB(modret) [Attribute retrieve $class $name]
    checkbutton .qmod.r -var PGB(modret) -text "Retrieve by default" \
	-anchor w 
    pack .qmod.com -fill x -in .qmod.top -padx 5m -pady 2m
    pack .qmod.l .qmod.r -fill x -in .qmod.top -padx 5m -side left -pady 2m
    frame .qmod.mid
    label .qmod.cl -text "Choices (one per line):"
    text .qmod.c -yscroll ".qmod.ys set" -width 40 -height 10 \
	-relief sunken -bd 2
    scrollbar .qmod.ys -command ".qmod.c yview"
    pack .qmod.cl -in .qmod.mid -fill x
    pack .qmod.ys -side right -in .qmod.mid -fill y 
    pack .qmod.c -in .qmod.mid -fill both
    .qmod.c insert 0.0 [join [Attribute choice $class $name] \n]

    frame .qmod.bottom
    Tku_Button .qmod.ok OK "QueryModOptions:OK $frame $class $name" "" 1
    Tku_Button .qmod.cancel Cancel {destroy .qmod}
    pack .qmod.ok .qmod.cancel -side left -padx 5m -pady 5m \
	-in .qmod.bottom
    pack .qmod.top .qmod.mid .qmod.bottom -fill x
    Tku_SetTabbing .qmod.com.e .qmod.l.e .qmod.c .qmod.ok.b .qmod.cancel.b
    Tku_Center .qmod
    tkwait visibilit .qmod.com.e
    focus .qmod.com.e
    tkwait window .qmod
}

proc QueryModOptions:OK {frame class name} {
    global PGB

    set len [.qmod.l.e get]
    if ![regexp {^[0-9]+$} $len] {
	Tku_Error "Invalid retrieve length - must be numeric."
	focus .qmod.l.e
	return 0
    }
    set comm [.qmod.com.e get]
    set ret $PGB(modret)
    set meta "$comm:retrieve=$ret:length=$len"
    set edited [string trim [.qmod.c get 0.0 end]]
    if {$edited != ""} {append meta :[join [split $edited \n] :]}
    if ![PGBSupportUpdate $class $name $meta] {return 0}
    Attribute resetmeta $class $name $meta
    QueryOptionsPullDown $frame $class $name
    set PGB(ret,$class,$name) $ret
    destroy .qmod
    return 1
}

# ---------------------------------------------------------------------------
# GraphQuery -    Get latlon info using IDL
# ---------------------------------------------------------------------------
proc GraphQuery {} {
    global PGB
    global coord_1 coord_2

	source /home/src/base/local/tk/pgb3.0/lib/graphical_query.tcl
	open_idl .idl
	graphical_query .idl
    loop i 0 $PGB(queryNum) {
	set text [.q.c.f.$i.r configure -text] 
	if [string match *RS_beginloc* $text] {
		.q.c.f.$i.entry.e delete 0 end
		.q.c.f.$i.entry.e insert 0 $coord_1
	}
	if [string match *RS_endloc* $text] {
		.q.c.f.$i.entry.e delete 0 end
		.q.c.f.$i.entry.e insert 0 $coord_2
	}
    }

} 
# ---------------------------------------------------------------------------
# QueryClear -    clear all the query entry boxes
# ---------------------------------------------------------------------------
proc QueryClear {} {
    global PGB

    loop i 0 $PGB(queryNum) {
	.q.c.f.$i.entry.e delete 0 end
	if [winfo exists .q.c.f.n$i] {destroy .q.c.f.n$i}
    }
    QueryResize
    focus .q.c.f.0.entry.e
} 

# ---------------------------------------------------------------------------
# QueryAppend -    add an instance to the class being queried
# ---------------------------------------------------------------------------
proc QueryAppend {} {
    global PGB

    set class $PGB(queryClasses)
    # build up the Postquel APPEND command, casting data to its type
    set natts [Attribute num $class]
    set nulls 0
    for {set i 0} {$i < $natts} {incr i 1} {
        set name $PGB(query,name,$i) 
	if {$name == "oid"} continue
        set type [Attribute type $class $name]
        set data [.q.c.f.$i.entry.e get]
                if {$data == {}} {
            incr nulls
            set nullfield $name
        }
        if ![info exists enter_query] {
	    set enter_query "append $class (\n\t$name=\"$data\"::$type"
        } else {
            append enter_query ",\n\t$name=\"$data\"::$type"
        }
    }
    append enter_query "\n)"

    # warn the user if they forgot to enter some data
    case $nulls in {
        0    {}
        1    {if ![Tku_Confirm "$nullfield is null, continue?"] return}
        default    {if ![Tku_Confirm "Some fields are null, continue?"] return}
    }

    # send the Postquel
    if ![PQ exec $enter_query] {
       PQ errbox "APPEND Failed"
       return 0
    }

    # clear the entry boxes and focus on the first entry box again
    QueryClear
    focus .q.c.f.0.entry.e
    return 1
} 

# ---------------------------------------------------------------------------
# QueryRetrieve -    perform a qualified search
# ---------------------------------------------------------------------------
proc QueryRetrieve {} {
    global PGB 

    # make sure the user wants at least a few attributes retrieved
    if {[set ret [QueryAttsToRet]] == ""} return

    set retrieve retrieve
    if $PGB(query,unique) {append retrieve " unique"}

    # build up the Postquel RETRIEVE command
    loop i 0 $PGB(queryNum) {
	if {[set qual [QueryGetQual $i]] != ""} {
		set andor [lindex $qual 0]
		lappend quals($andor) $qual
	}
    	set nframe .q.c.f.n$i
	if [winfo exists $nframe] {
	    foreach b [winfo children $nframe] {
		set qual [lindex [$b conf -text] 4]
		set andor [lindex $qual 0]
		lappend quals($andor) $qual
	    }
	}
    }
#    foreach qual [Tku_ListboxDump .q.qs.listbox] {
#	set andor [lindex $qual 0]
#	lappend quals($andor) $qual
#    }
	
    if ![info exists quals] {
	Tku_Notice "Please qualify your search" 
	return 0
    }
    if ![info exists quals(and)] {
	set quals(and) [list [lvarpop quals(or)]]
    }
    foreach qual $quals(and) {
	if ![info exists lookup] {
	    regsub {^and|or} $qual where qual
            set lookup "$retrieve ($ret)\n\t$qual"
	} else {
            append lookup "\n\t$qual"
	}
    }
    if [info exists quals(or)] {
	foreach qual $quals(or) {append lookup "\n\t$qual"}
    }

    # add sorting if needed
    if {$PGB(querySortBy) != -1} {
	append lookup "\nsort by $PGB(querySortBy)"
    }

    # send query and view results
    if ![PQ exec $lookup] {
        PQ errbox "Retrieve failed"
        return 0
    }
    QRes
}

# ---------------------------------------------------------------------------
# QueryRetrieveAll - retrieve all instances in a class
# ---------------------------------------------------------------------------
proc QueryRetrieveAll {} {
    global PGB

    # make sure the user wants at least a few attributes retrieved
    if {[set ret [QueryAttsToRet]] == ""} return

    # create the RETRIEVE command with sorting if needed
    set cmd retrieve
    if $PGB(query,unique) {append cmd " unique"}
    append cmd "($ret)"
    if {$PGB(querySortBy) != -1} {
	append cmd "\nsort by $PGB(querySortBy)"
    }

    # send the Postquel and show the results
    if ![PQ exec "$cmd"] {
        PQ errbox "Retrieve all failed"
        return 0
    }
    QRes
}

# ---------------------------------------------------------------------------
# QueryAttsToRet - create a list of attributes for a RETRIEVE command
# ---------------------------------------------------------------------------
proc QueryAttsToRet {} {
    global PGB 


    # append the attributes to retrieve for each class
    foreach class $PGB(queryClasses) {

      set all 1
      foreach name [Attribute names $class] {
	if $PGB(ret,$class,$name) {
	    lappend ret($class) "$class.$name"
	} else {
	    set all 0
	}
      }

      # use .all if all attributes in this class are selected
      if $all {set ret($class) "$class.all,$class.oid"}
     
      if [info exists ret($class)] {
        # always retrieve the oid for updateable queries
        if $PGB(queryUpdatable) {
	    if !$PGB(ret,$class,oid) {lappend ret($class) "$class.oid"}
	}
	set ret($class) [join $ret($class) ,]
      }

    }

    if ![info exists ret] {
	Tku_Notice "Please select one or more attributes."
	return ""
    }

    foreach class [array names ret] {lappend allret $ret($class)}
    return [join $allret ,]
}

# ---------------------------------------------------------------------------
# QueryClose -    closes the query and query results windows
# ---------------------------------------------------------------------------
proc QueryClose {} {
    global PGB 

    catch {destroy .qres}
    catch {destroy .q}

    # exit if pgbrowse was started with just the query tool
    if ![winfo exists .dbs] CloseBrowser
}

# ---------------------------------------------------------------------------
# QRes - show the results of a query in a spreadsheet like set of listboxes
# ---------------------------------------------------------------------------
proc QRes {} {
    global PGB

    if ![PQ nlines] {
        Tku_Notice {Nothing retrieved.}
        return 0
    }

    # make sure the user wants to wait if PQ has returned a lot of data
    if {[expr [PQ nlines] * [PQ ncols]] > 100} {
    	if ![Tku_Confirm "[PQ nlines] rows in [PQ ncols] columns returned." View Cancel] {return 0}
    }

    set PGB(qr,ncols) [set ncol [PQ ncols]]
    set PGB(qr,nlines) [set nlines [PQ nlines]]

    Tku_Working "Extracting $PGB(qr,nlines) rows..."

    if [winfo exists .qres] {destroy .qres}
    toplevel .qres
    wm title .qres "Query Results"
    wm minsize .qres 100 300
    wm geom .qres +300+300
    wm withdraw .qres

    Tku_MenuBar .qres.menu \
	{About {This window shows the results of your query.  To resize
	columns, drag their right side border.  You can select entries
	by pressing MB1 and dragging.  To extend the selection or add a
	column, press Shift-MB1.  Selected rows and/or columns can then
	be processed with user-written Actions (see the README file in
	the $PGBROWSE/action directory for more information on
	Actions).  Also, if you're working with a single class you can
	edit or delete the selected rows.}}

    Tku_MenuButton .qres.menu "File -under 0" \
        "command -label Save -command QResSave -accel C-S -under 0" \
	{Saves columns and lines you have selected to a file.} \
        "command -label Print -command QResPrint -accel C-P -under 0" \
	{Prints columns and lines you have selected.} \
        "command -label Close -command {destroy .qres} -accel C-W -under 0" \
	{Closes the results of your query.}

    pack .qres.menu -fill x

    Tku_StatusBar .qres.info "Query Results"
    pack .qres.info -fill x


    if $PGB(queryUpdatable) {set cmd "QResEdit .qres.sheet"} \
    else {set cmd {Tku_Notice "This output is read only."}}
    loop n 0 $PGB(qr,ncols) {
	set name [string trim [PQ getname $n]]
	set max 0
	foreach class $PGB(queryClasses) {
	    if ![catch {set n [Attribute length $class $name]}] {
		if $n>$max {set max $n}
	    }
	}
	lappend colinfo [list $name $max $cmd]
    }
    Tku_Sheet create .qres.sheet $PGB(qr,ncols) $PGB(qr,nlines) $colinfo
    loop n 0 $PGB(qr,ncols) {Tku_Sheet putcol .qres.sheet $n [PQ getcol $n]}
    pack .qres.sheet -fill both -expand 1

    # add the edit and delete buttons at the bottom
    Tku_Button .qres.all All "Tku_Sheet selectall .qres.sheet" \
	"Selects all lines"
    Tku_Button .qres.clear Clear "Tku_Sheet selectclear .qres.sheet" \
	"Clears the selection"
    lappend tablist .qres.all.b .qres.clear.b
    pack .qres.all .qres.clear -side left -padx 5m -pady 2m

    # add the edit and delete buttons at the bottom
    if $PGB(queryUpdatable) {
    	Tku_Button .qres.edit "Edit..." "QResEdit .qres.sheet" \
		"Edits selected instances"
    	Tku_Button .qres.delete "Delete..." "QResDelete .qres.sheet" \
	    "Deletes selected instances"
    	pack .qres.edit .qres.delete -side left -padx 5m -pady 2m
	lappend tablist .qres.edit.b .qres.delete.b
    } else {
	pack [label .qres.ro -text "Read Only"] -side left -padx 5m -pady 2m
    }

    # add the action button
    ActionButton .qres.menu 

    wm deiconify .qres
    eval Tku_SetTabbing $tablist
    Tku_DefFocus .qres.all.b
    Tku_Waiting
}

# ---------------------------------------------------------------------------
# QResDelete - delete a line from the database
# ---------------------------------------------------------------------------
proc QResDelete {s} {
    global PGB

    set class $PGB(queryClasses)

    if {[set insts [Tku_Sheet cur $s]] == ""} return
    if [Tku_Confirm "Delete the selected instances?" No Yes] return

    # delete instances 10 at a time
    set oidbox $s.c.[expr $PGB(qr,ncols) - 1]
    set n [llength $insts]
    set i 0
    set j 0
    while {$i < $n} {
	incr i 9
	if {$i >= $n} {
 	    set ten [lrange $insts $j end]
	} else {
	    set ten [lrange $insts $j $i]
	}
	set j [incr i]

    	foreach inst $ten {
            set oid "\"[$oidbox get $inst]\"::oid"
            if ![info exists cmd] {
            	set cmd "delete $class\n\twhere $class.oid = $oid"
            } else {
            	append cmd "\n\tor    $class.oid = $oid"
            }
    	}

    	if ![PQ exec $cmd] {
            PQ errbox "Delete failed"
            return 0
    	}
    	unset cmd
    }

    # delete the lines from the list boxes and update line number column
    foreach i [lreverse $insts] {Tku_Sheet delete $s $i}
}

# ---------------------------------------------------------------------------
# QResEdit - start editing one or more instances
# ---------------------------------------------------------------------------
proc QResEdit {s} {
    global PGB

    if [winfo exists .qed] return
    if {[set toedit [Tku_Sheet currentsel $s]] == ""} return

    # if editing more than one line, let the user cancel mid way through
    set PGB(CancelAll) 0; set PGB(MultiEdit) 0
    if {[llength $toedit] != 1} {set PGB(MultiEdit) 1}
    foreach i $toedit {if !$PGB(CancelAll) {QResEditOne $s $i}}
}

# ---------------------------------------------------------------------------
# QResEditOne - edit a single instance
# ---------------------------------------------------------------------------
proc QResEditOne {s i} {
    global PGB

    set class $PGB(queryClasses)
    set w [toplevel .qed]
    wm title $w "Edit Instance #$i"
    wm iconname $w "Edit #$i"

    canvas $w.c -yscroll "$w.ys set"
    scrollbar $w.ys -command "$w.c yview"
    frame $w.c.f
    $w.c create window 0 0 -anchor nw -window $w.c.f

    set ndatacol [expr $PGB(qr,ncols) - 1]
    set max [Attribute maxline $class]
    loop c 0 $ndatacol {
	set box $w.c.f.$c
	set lastbox $w.c.f.[expr $c-1]
	set name [Tku_Sheet name $s $c]
	Tku_EntryBox $box [Attribute def $class $name] \
		[Tku_Sheet get $s $c $i] -width 50
	$box.label conf -width $max -anchor e
        if $c {bind $lastbox.e <Return> "focus $box.e"}
        lappend tablist $box.e
	pack $box -anchor w -fill x
    }
    bind $lastbox.e <Return> "focus $w.ok.b"
    frame $w.bs
    Tku_Button $w.ok Commit "QResEditOk $s $i" \
	"Commit changes to this instances" 1
    Tku_Button $w.cancel Cancel "destroy $w" "Cancel editing of this instances"
    pack $w.ok $w.cancel -in $w.bs -side left -padx 5m -pady 5m
    lappend tablist $w.ok.b $w.cancel.b

    # add a button to cancel multiple instance editing
    if $PGB(MultiEdit) {
        Tku_Button $w.canall "Cancel All" \
	    "set PGB(CancelAll) 1; destroy .qed" \
	    "Cancel editing this and all other selected instances"
 	pack $w.canall -in $w.bs -padx 5m -side left
        lappend tablist $w.canall.b
    }

    eval Tku_SetTabbing $tablist
    pack $w.bs -side bottom -fill x
    pack $w.ys -fill y -side right
    pack $w.c -fill both -expand 1
    wm withdraw $w
    update
    set wid [winfo reqwidth $w.c.f]
    set th [set h [winfo reqheight $w.c.f]]
    set maxh [expr [winfo screenheight $w]*.8]
    if {$th > $maxh} {set th $maxh} else {pack forget $w.ys}
    $w.c conf -scrollregion "0 0 $wid $h" -width $wid -height $th
    wm minsize $w 100 100
    Tku_Center $w
    tkwait visibility $w.c.f.0.e
    focus $w.c.f.0.e
    tkwait window $w
}

# ---------------------------------------------------------------------------
# QResEditOk - submit REPLACE command
# ---------------------------------------------------------------------------
proc QResEditOk {s e} {
    global PGB

    set class $PGB(queryClasses)

    set natts $PGB(qr,ncols)
    incr natts -1
    set oid [Tku_Sheet get $s $natts $e]

    set nulls 0
    loop col 0 $natts {
	set name [Tku_Sheet name $s $col]
	set data [.qed.c.f.$col.e get]
	if {$data == [Tku_Sheet get $s $col $e]} continue
	if {$data == "(null)"} {set data {}}
	if {$data == {}} {
	    incr nulls 1
	    set nullfield $name
	} elseif ![info exists enter_query] {
            set enter_query "replace $class ($name=\"$data\""
	} else {
            append enter_query ",\n\t$name=\"$data\""
        }
    }
    if ![info exists enter_query] {
	destroy .qed
	Tku_Notice "No changes made to instance #$e"
	return
    }
    case $nulls in {
        0    {}
        1    {if ![Tku_Confirm "$nullfield is null, continue?"] return}
        default    {if ![Tku_Confirm "Some fields are null, continue?"] return}
    }
    append enter_query "\n) where $class.oid = \"$oid\"::oid"
    if ![PQ exec $enter_query] {
        PQ errbox "Replace Failed"
        return 0
    }
    loop col 0 $natts {
	Tku_Sheet replace $s $col $e [.qed.c.f.$col.e get]
    }
    destroy .qed
}

# ---------------------------------------------------------------------------
# QResAllBtn - select all/none instances
# ---------------------------------------------------------------------------
proc QResAllBtn {} {
    global PGB
    if $PGB(all) {
	Tku_Sheet selectall $s
    } else {
	Tku_Sheet selectclear $s
    }
}

proc QResSave {} {Tku_Sheet save .qres.sheet}
proc QResPrint {} {Tku_Sheet print .qres.sheet}
proc ActionColNum {name {allok 1}} {Tku_Sheet:ColNum .qres.sheet $name $allok}
proc ActionGetCol {n {allok 1}} {Tku_Sheet:GetCol .qres.sheet $n $allok}
proc ActionFields {} {Tku_Sheet:ColNames .qres.sheet}
